/* SubStation alpha support */

#include "config.h"

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#include <stdio.h>
#include <string.h>

#include "libsub_private.h"


/*                name                identify     parse      valid exts. */
#define _FORMAT_ {"SubStation Alpha",ssa_identify,ssa_parse,{"ssa",0,0,0}},


enum
{
  FIELD_IGNORE = 0,
  // Style

  FIELD_ALIGN,
  FIELD_NAME,    FIELD_FONTNAME,   FIELD_FONTSIZE,
  FIELD_COLOR1,  FIELD_BACKCOLOR, FIELD_BOLD,       FIELD_ITALIC, 
  FIELD_OUTLINE, FIELD_SHADOW,     FIELD_MARGINL,  FIELD_MARGINR,
  FIELD_MARGINV, FIELD_ALPHALEVEL, FIELD_ENCODING, FIELD_COLOR2,
  FIELD_COLOR3,

  // Event
  FIELD_START,  FIELD_END,  FIELD_STYLE,  FIELD_TEXT,

  FIELD_LAST,
};

static const struct
{
  const char *name;
  const int id;
}
fields[] =
{
  { "alignment",        FIELD_ALIGN      },
  { "alphalevel",       FIELD_ALPHALEVEL },
  { "backcolour",       FIELD_BACKCOLOR  },
  { "bold",             FIELD_BOLD       },
  { "encoding",         FIELD_ENCODING   },
  { "fontname",         FIELD_FONTNAME   },
  { "fontsize",         FIELD_FONTSIZE   },
  { "italic",           FIELD_ITALIC     },
  { "marginl",          FIELD_MARGINL    },
  { "marginr",          FIELD_MARGINR    },
  { "marginv",          FIELD_MARGINV    },
  { "name",             FIELD_NAME       },
  { "outline",          FIELD_OUTLINE    },
  { "primarycolour",    FIELD_COLOR1     },
  { "secondarycolour",  FIELD_COLOR2     },
  { "tertiarycolour",   FIELD_COLOR3     },
  { "shadow",           FIELD_SHADOW     },

  { "start", FIELD_START },
  { "end",   FIELD_END   },
  { "style", FIELD_STYLE },
  { "text",  FIELD_TEXT  },
};

#define NFIELDS NELEM(fields)

#define MAX_SSA_STYLES 128

struct style
{
  char *name;
  int   at;
  int   ml, mr, mv;

  char *font;
  int   size;
  int   color, color2, color3, scolor, ocolor;
  int   flags;
};

static double time_adjust = 100.0;
static int height = 550;
static struct ssa_state
{
  struct style *styles[MAX_SSA_STYLES];
  int next_style;
  int stage;
  int hard;
  unsigned char fields[FIELD_LAST];
} state;
  

static void ssa_cleanup()
{
  int i;
  for( i = 0; i<state.next_style; i++ )
  {
    if( state.styles[i]->name )
      xfree( state.styles[i]->name );
    if( state.styles[i]->font )
      xfree( state.styles[i]->font );
    xfree( state.styles[i] );
  }
  state.next_style = 0;
}
#include <ctype.h>

static int ssa_read_line( libsubFILE *fd, struct subtitle *current )
/* Overly ambition function, yes, but it's rather handy when you write
 * your own .sub-files to be able to include and exclude any fields at
 * will, and use any order. :-) */
{
  int i, j;
  unsigned char line[1001];
  char *part;
  char  *lp;
  int cnt;
  char *split[64];
  struct style style;
  struct {
    int flags;
    int nflags;
    int color;
    int delay;
    char *font;
  } override;

  memset( &override, 0, sizeof( override ) );
  while( fd->fgets( line, 1000, fd ) )
  {
    line[1000]=0;
    int l = strlen( line )-1;


    while( l >= 0 && isspace(line[l]) )
      l--;

    line[l+1] = 0;
    
    //    fprintf(stderr,"%s <%d>\n", line, strlen(line));
    if( line[0] == '[' )
    {
      if( !strncmp( line, "[V4 Styles]", 11 ) )
	state.stage = 1;
      else if( !strncmp( line, "[Events]", 8 ) )
	state.stage = 2;
      else
	state.stage = 0;

      continue;
    }
    
    //////// Actual parsing
    
    switch( state.stage )
    {
      case 0: // outside known section.
	sscanf( line, "PlayResY: %d", &height );
	sscanf( line, "Timer: %f", &time_adjust );
	break;

      case 1: // styles
	if( state.next_style == MAX_SSA_STYLES )
	  continue;
	switch( line[0] )
	{
	  case 'f':
	  case 'F': // Format:
	    cnt = 0;
	    if( !strncasecmp( line, "format:", 7 ) )
	      lp = line+8;
	    else
	      continue;
	    memset( state.fields, 99, sizeof( state.fields ) );
	    while( part = strsep( &lp, "," ) )
	    {
	      while( *part==' ' )
		part++;
	      for( i = 0; i<(int)NFIELDS; i++ )
		if( fields[i].name &&
		    !strncasecmp( fields[i].name, part, strlen( fields[i].name ) ) )
		  state.fields[ fields[i].id ] = cnt;
	      cnt++;
	    }
	    break;

	  case 's':
	  case 'S': // Style:
	    if( !strncasecmp( line, "style:", 6 ) )
	      lp = line+7;
	    else
	      continue;	
	    if( state.fields[ FIELD_NAME ] == 99 )
	    {
	      fprintf( stderr, "No 'Name' field in the style. Cannot parse\n");
	      continue;
	    }

	    cnt = 0;
	    memset( split, 0, sizeof( split ) );
	    while( part = strsep( &lp, "," ) )
	    {
	      while( *part==' ' ) part++;
	      split[ cnt++ ] = part;
	    }

	    if( lp = split[ state.fields[ FIELD_NAME ] ] )
	      // create a new style.
	    {
	      cnt = state.next_style++;
	      state.styles[cnt] = xalloc( sizeof( struct style ) );
	      memset( state.styles[cnt], 0, sizeof( struct style ) );
	      state.styles[cnt]->name = xstrdup( lp );
	      state.styles[cnt]->at = 0;
	    }
	    else
	      continue;

#define FIELD( X ) (state.fields[ X ] && (lp = split[ state.fields[ X ] ] ))
#define IFIELD( X ) (state.fields[ X ] && (lp = split[ state.fields[ X ] ] ) && ( (j=atoi(lp)), 1))

	    if( IFIELD( FIELD_ALIGN ) )
	    {
	      if( (j & 3) == ALIGN_LEFT) // left.
		j = (j&12) | 1;
	      if( (j & 3) == 2) // center.
		j = (j&12);
	      else              // right
		j = (j&12) | ALIGN_RIGHT;
	      state.styles[ cnt ]->at = j;
	    }

	    if( FIELD( FIELD_FONTNAME ) )
	      state.styles[ cnt ]->font = xstrdup( lp );

	    if( IFIELD( FIELD_FONTSIZE ) )
	      state.styles[ cnt ]->size = j;

	    if( IFIELD( FIELD_MARGINL ) ) state.styles[ cnt ]->ml = j;
	    if( IFIELD( FIELD_MARGINR ) ) state.styles[ cnt ]->mr = j;
	    if( IFIELD( FIELD_MARGINV ) ) state.styles[ cnt ]->mv = j;

	    // In file: B G R
	    // Wanted:  R G B
#define FIX_COLOR( X ) (((X>>16)&255) | (X&(255<<8)) | ((X&255)<<16))
	    
	    if( IFIELD( FIELD_COLOR1 ) )
	      state.styles[ cnt ]->color = FIX_COLOR( j );

	    if( IFIELD( FIELD_COLOR2 ) )
	      state.styles[ cnt ]->color2 = FIX_COLOR( j );

	    if( IFIELD( FIELD_COLOR3 ) )
	      state.styles[ cnt ]->color3 = FIX_COLOR( j );

	    if( IFIELD( FIELD_BACKCOLOR ) )
	      state.styles[ cnt ]->scolor =
		state.styles[ cnt ]->ocolor = FIX_COLOR( j );

	    if( IFIELD( FIELD_BOLD )  && ( j == -1 ) )
	      state.styles[ cnt ]->flags |= FONT_BOLD;
	    
	    if( IFIELD( FIELD_ITALIC ) &&  (j == -1) )
	      state.styles[ cnt ]->flags |= FONT_ITALIC;

	    if( IFIELD( FIELD_OUTLINE )  && j )
	      state.styles[ cnt ]->flags |= FONT_OUTLINE;
	    
	    if( IFIELD( FIELD_SHADOW )  && j )
	      state.styles[ cnt ]->flags |= FONT_SHADOW;

	    break;
	}
	break;

      case 2: // events
	switch( line[0] )
	{
	  case 'f': // Format:
	  case 'F': // format:
	    cnt = 0;
	    if( !strncasecmp( line, "format:", 7 ) )
	      lp = line+8;
	    else
	      continue;
	    memset( state.fields, 99, sizeof( state.fields ) );
	    while( part = strsep( &lp, "," ) )
	    {
	      while( *part==' ' )
		part++;
	      for( i = 0; i<(int)NFIELDS; i++ )
		if( fields[i].name && !strncasecmp( fields[i].name, part, strlen( fields[i].name ) ) )
		  state.fields[ fields[i].id ] = cnt;
	      cnt++;
	    }
	    break;

	  case 'D': // Dialogue:
	  case 'd': // dialogue:
	    if( !strncasecmp( line, "dialogue:", 9 ) )
	      lp = line+10;
	    else
	      continue;

	    cnt = 0;
	    memset( split, 0, sizeof( split ) );

	    while( part = strsep( &lp, "," ) )
	      split[ cnt++ ] = part;

	    // The fields we must have: Start, End, Text

	    if( (state.fields[ FIELD_START ] == 99) ||
		(state.fields[ FIELD_END ] == 99)   ||
		(state.fields[ FIELD_TEXT ] == 99) )
	    {
	      fprintf( stderr, "Cannot parse subtitle data\n"
		       "There must be at least START, END and TEXT fields\n");
	      continue;
	    }

	    if( state.fields[ FIELD_STYLE ] != 99 &&
		(lp = split[ state.fields[ FIELD_STYLE ] ] ) )
	    {
	      if( lp[0] == '*' )
		lp++;
	      for( i = 0; i<state.next_style; i++ )
		if( !strcasecmp( lp, state.styles[i]->name ) )
		{
		  style = *state.styles[i];
		  break;
		}
	      if( i == state.next_style )
	      {
		fprintf( stderr, "No style found for '%s'!\n", lp );
		memset( &style, 0, sizeof( struct style ) );
		style.font = "Arial";
		style.size = 24;
		style.color = 0xffff00;
		style.flags = FONT_BOLD;
	      }
	    }

	    {
	      int h, m, s, hu;
	      lp = split[ state.fields[ FIELD_START ] ];
	      if( !lp )
	      {
		fprintf( stderr, "Missing Start field\n");
		continue;
	      }

	      sscanf( lp, "%d:%d:%d.%d", &h, &m, &s, &hu );
	      current->start = ((h*360000 + m*6000 + s*100 + hu)
				* ((int)time_adjust)) / 100;
	      lp = split[ state.fields[ FIELD_END ] ];
	      if( !lp )
	      {
		fprintf( stderr, "Missing End field\n");
		continue;
	      }

	      sscanf( lp, "%d:%d:%d.%d", &h, &m, &s, &hu );
	      current->end = ((h*360000 + m*6000 + s*100 + hu)
			      * ((int)time_adjust)) / 100;
	    }

	    // The text with it's formatting information. Color and
	    // font style is ignored.
	    lp = split[ state.fields[ FIELD_TEXT ] ];

	    if( !lp ) // empty line...
	      continue;

	    // Now.. All fields from FIELD_TEXT onwards are FIELD_TEXT
	    // (according to spec.). So, lets replace the \0's with
	    // ,'s again. This could be optimized.

	    for( i = state.fields[ FIELD_TEXT ]+1; split[i]; i++ )
	      split[i][-1]=',';


	    // Transfer from style to current.

	    /* FontSize to PixelSize. Very aproximate. */
#define FS_TO_PS(X) ((X)*0.71)

	    
	    current->size   =
	      FS_TO_PS((float)((double)style.size / height ));
	    current->flags  = style.flags;
	    current->color  = style.color;
	    current->scolor = style.scolor;
	    current->ocolor = style.ocolor;
	    current->at     = style.at;
	    

	    if( lp[0] == '{' )
	    {
	      for( i = 0; i<(int)strlen( lp ); i++ )
	      {
		if( lp[i] == '}' )
		{
		  lp += i+1;
		  break;
		}
		if( lp[i] == '\\' )
		{
		  if( lp[i+i] == 'k' )
		      // karaoke delay.
		  {
		    override.delay = atoi( lp + ++i );
		    override.color = style.color2;
		    override.flags |= FONT_COLOR;
		    override.nflags &= ~FONT_COLOR;
		  }
		  if( lp[i+1] == 'a')
		  {
		    int a = atoi( lp+i+2 );
		    if( (a & 3) == 1) // left.
		      current->at = (a&12) | 1;
		    if( (a & 3) == 2) // center.
		      current->at = (a&12);
		    else              // right
		      current->at = (a&12) | 2;
		    i++;
		  }
		  else if( lp[i+1] == 'b' )
		  {
		    if( lp[i+2] == '0' )
		      current->flags &= ~FONT_BOLD;
		    else if( lp[i+2] == '1' )
		      current->flags |= FONT_BOLD;
		  }
		  else if( lp[i+1] == 'i' )
		  {
		    if( lp[i+2] == '0' )
		      current->flags &= ~FONT_ITALIC;
		    else if( lp[i+2] == '1' )
		      current->flags |= FONT_ITALIC;
		  }
		  else if( lp[i+1] == 'f' && lp[i+2] == 'n')
		  {
		    if( override.font )
		      xfree( override.font );
		    override.font = lp+3;
		    for( i = i+3; lp[i] && lp[i] != '}'; i++ );
		    lp[i] = '\0';
		    override.font = xstrdup( override.font );
		    lp[i] = '}';
		    i--;
		  }
		  else if( lp[i+1] == 'f' && lp[i+2] == 's')
		    current->size=FS_TO_PS((float)((double)atoi(lp+i+3)/height));
		  else if( lp[i+1] == 'f' && lp[i+2] == 'e' )
		    ; // FIXME: font encoding
		  else if( lp[i+1] == 'c' && lp[i+2] == '&' && lp[i+3] == 'H' )
		  {
		    int c = 0;
		    sscanf( lp+i+4, "%x&", &c );
		    current->color = FIX_COLOR( c );
		  }
		}
	      }
	    }

	    split[ state.fields[FIELD_TEXT] ] = lp;

	    switch( current->at & 3 )
	    {
	      case 0:
		break;
	      case 1:
		// left
		if( IFIELD( FIELD_MARGINL ) && j )
		  current->hm = ((double)j)/(height*4/3.0);
		else
		  current->hm = ((double)style.ml)/(height*4/3.0);
		break;
	      case 2:
		// left
		if( IFIELD( FIELD_MARGINR ) && j )
		  current->hm = ((double)j)/(height*4/3.0);
		else
		  current->hm = ((double)style.ml)/(height*4/3.0);
		break;
	    }

	    switch( current->at>>2 )
	    {
	      case 1:
	      case 2:
		if( IFIELD( FIELD_MARGINV ) && j)
		  current->vm = ((double)j)/height;
		else
		  current->vm = ((double)style.mv)/height;
		break;
	    }


	    lp = split[ state.fields[FIELD_TEXT] ];

	    /* Get the parts. */

	    {
	      int l = strlen( lp ), state = 0, new = 1, p = 0;
	      int c = 0;
	      int allocated_size = 0;
	      char cl[1000];
	      struct part *cp = 0;
	      memset( cl, 0, 1000 );
	      current->ntext = 0;

	      for( i=0; i<l; i++ )
	      {
		if( !state )
		  /* outside {} */
		{
		  if( lp[i] == '{' )
		  {
		    state = 1;
		    continue;
		  }

		  if( new )
		  {
		    new = 0;

		    if( current->ntext >= allocated_size )
		    {
		      allocated_size += 10;
		      current->text = realloc( current->text,
					       allocated_size*
					       sizeof(current->text[0]) );
		      if( cp )
			cp = &current->text[ current->ntext-1 ];
		    }
		    
		    if( cp ) // last segment.
		      cp->str = sub_iso8859_to_unicode( cl );

		    memset( cl, 0, 1000 );
		    cp = &current->text[ current->ntext++ ];
		    memset( cp, 0, sizeof( *cp ) );
		    p = 0;
	    	    cp->font =
		      lsub_find_font( (override.font?override.font:style.font),
				      ((current->flags|override.flags)
				       & ~override.nflags ) );
		    cp->flags = override.flags;
		    cp->nflags = override.nflags;
		    cp->nl = 0;
		    cp->color = override.color;
		    cp->delay = override.delay;
		    cp->str = 0;
		  }

		  switch( lp[i] )
		  {
		    case '\t':
		    case ' ':
		      cl[ p++ ] = ' ';
		      new = 1;
		      continue;

		    case '\\':
		      i++;
		      switch( lp[i] )
		      {
			case '{':
			case '}':
			case '\\':
			  c = lp[i];
			  break;

			case 'n': // soft newline
			  new = 1;
/* 			  if( !state.hard ) */
/* 			  { */
			    cl[ p++ ] = ' ';
			    continue;
/* 			  } */
			case 'N': // hard newline
			  cp->nl = 1;
			  new = 1;
			  continue;
		      }
		    default:
		      c = lp[i];
		      break;
		  }
		  cl[ p++ ] = c;
		}
		else /* inside {} */
		{
		  if( lp[i] == '}' )
		  {
		    state = 0;
		    new = 1;
		    continue;
		  }
		  if( lp[i] == '\\' )
		  {
		    i++;
		    if( lp[i] == 'a')
		    {
		      ; // at cannot be parsed once the subtitle has
			// been positioned!
		    }
		    else if( lp[i] == 'b' )
		    {
		      i++;
		      if( lp[i] == '0' )
		      {
			override.nflags |= FONT_BOLD;
			override.flags &= ~FONT_BOLD;
		      }
		      else if( lp[i] == '1' )
		      {
			override.flags |= FONT_BOLD;
			override.nflags &= ~FONT_BOLD;
		      }
		    }
		    else if( lp[i] == 'i' )
		    {
		      i++;
		      if( lp[i] == '0' )
		      {
			override.nflags |= FONT_ITALIC;
			override.flags &= ~FONT_ITALIC;
		      }
		      else if( lp[i] == '1' )
		      {
			override.flags |= FONT_ITALIC;
			override.nflags &= ~FONT_ITALIC;
		      }
		    }
		    else if( lp[i] == 'f' && lp[++i] == 'n')
		    {
		      if( override.font )
			xfree( override.font );
		      override.font = lp+1;
		      for( i = i+1; lp[i] && lp[i] != '}'; i++ );
		      lp[i] = '\0';
		      override.font = xstrdup( override.font );
		      lp[i] = '}';
		      i--;
		    }
		    else if( lp[i] == 'f' && lp[++i] == 's')
		      ;// FIXME: Fontsize ignored here.
   		       //  I have yet to find a script with more than
   		       //  one fontsize on one line.. The layout would
   		       //  be a tad harder.
		    else if( lp[i] == 'f' && lp[++i] == 'e' )
		      ; // FIXME: font encoding
		    else if( lp[i] == 'c' && lp[++i] == '&' &&
			     lp[++i] == 'H' )
		    {
		      int c = 0;
		      i++;
		      sscanf( lp+i, "%x&", &c );
		      override.color = FIX_COLOR( c );
		      override.nflags &= ~FONT_COLOR;
		      override.flags |= FONT_COLOR;
		    }
		    else if( lp[i] == 'r' )
		      // reset.
		    {
		      override.flags = override.nflags = override.color = 0;
		    }
		    else if( lp[i] == 'k' )
		      // karaoke delay.
		    {
		      override.delay = atoi( lp + ++i );
		      override.color = style.color2;
		      override.flags |= FONT_COLOR;
		      override.nflags &= ~FONT_COLOR;
		    }
		  }
		}
	      }
	      if( cp )
	      {
		cp->str = sub_iso8859_to_unicode( cl );
		cp->nl = 1;
	      }
	    }
	    if( override.font )
	      xfree( override.font );
	    return 1;
	}
    }
  }
  return 0;
}

int ssa_identify( libsubFILE *f )
{
  char line[80];
  line[79] = 0;
  f->rewind( f );
  while( f->fgets( line, 79, f ) )
    if( !strncmp( line, "[Events]", 8 ) )
      return 1;
  return 0;
}

#define HAS_STYLE
#define HAVE_KARAOKE
#define LFUN ssa_read_line
#define PFUN ssa_parse
#define CLEANUP() ssa_cleanup()
#include "sub_parse_function.h"
